# /r2/sketch_talk_classes.rb

# SketchTalk Classes ------------------------------------------

# typing savers

class GT < Geom::Transformation
end

class P3d < Geom::Point3d
end

class V3d < Geom::Vector3d
end

# other classes

class Box
=begin
A box is a rectangular parallelepiped: six faces, all rectangular, all meeting at right angles. A box is defined by near and far corners of a rectangle plus a distance to PushPull the rectangle, creating the box shape. If the rectangle is orthogonal, a positive PushPull is done toward the positive end of the perpendicular axis. All six faces of the box will face the outside of the box.
=end
    attr_reader :rectangle, :pushpull_distance

    def initialize( near, far, pushpull_distance )     
        @rectangle = Rectangle.new( near, far )
        @pushpull_distance = pushpull_distance
    end
    
    def draw
        @rectangle.draw()
        @rectangle.face().pushpull( pushpull_distance )
    end
    
end # of class Box

class Circle
    attr_reader :nsides, :center, :radius, :normal, :edges, :face, :group
    
    def initialize( nsides, center, radius, plane_or_normal )
        @nsides = nsides
        @center = center
        @radius = radius
        @plane_or_normal = plane_or_normal
        @normal = make_normal( plane_or_normal )
    end # of initialize()
    
    def draw( ents )
        @group = ents.add_group()
        ents = @group.entities()
        @edges = ents.add_circle( @center, @normal, @radius, @nsides )
        @face = ents.add_face( @edges )
        orient( @face )
        ents.add_face( @edges )
        Sketchup.active_model().selection().add( @group )
        return @group
    end # of draw()
    
    def inspect()
        ret = self.to_s() + @nsides.to_s() + ' sides, center=' + (center.inspect) + ', radius='+@radius.to_s()
    end # of inspect()
        
    
end # of class Circle

class Column
=begin
A Column is a Circle PushPulled to a given height.
=end
    attr_reader :circle, :height

    def initialize( nsides, center, radius, plane_or_normal, height )
        @nsides = nsides
        @center = center
        @radius = radius
        @plane_or_normal = plane_or_normal
        @normal = make_normal( @plane_or_normal )
        @height = height
    end # of initialize()
    
    def draw( ents )
        @circle = Circle.new( @nsides, @center, @radius, @normal ) 
            # returns a Circle
        @circle.draw( ents )
        @circle.face().pushpull( @height )
    end # of draw()

end # of class Column

class Doughnut
=begin
By hand: drag a Rectangle from near to far. oFFset inside the rectangle. Delete the face in the center. PushPull the remaining face.
=end
    attr_reader :rectangle, :pushpull_distance, :thickness, :group

    def initialize( near, far, pushpull_distance, thickness )
=begin
Parameters near and far are [r,g,b] arrays. The thickness is the distance from the outside to the hole.
=end    
        
        @rectangle = Rectangle.new( near, far )
        @pushpull_distance = pushpull_distance
        @thickness = thickness
        
    end # of initialize()
    
    def draw()
        @rectangle.draw()
        inner_rec = inset( @rectangle.near, @rectangle.far, @thickness )
        inner_rec.draw( @rectangle.group().entities() )
        inner_rec.group().explode()
        inner_rec.face().erase!()
        @rectangle.face().pushpull( @pushpull_distance )
    end # of draw()

    def inset( near, far, thickness )
=begin
        Computes new near/far pair, oFFset inside by thickness.
        This is messy as there is no guarantee that the coordinates
        of "near" are smaller than the coordinates of "far".
=end
        near2 = []
        far2 = []
        
        # First, assign smaller+thickness to near2, larger-thickness to far2.
        for i in 0..2
            arr = ins( near[i], far[i] )
            near2[i] = arr[0]
            far2[i] = arr[1]
        end

        # The coordinates of the perpendicular value V are now V+thickness and
        # V-thickness. ([0,0,0],[100,100,0] is now [8,8,8], [92,92,-8]).
        # Adjust them back to V and V.
        unless @plane == 'no'
            case @rectangle.plane()
                when 'rg' then i = 2
                when 'rb' then i = 1
                when 'gb' then i = 0
            end
            
            near2[i] = far2[i] = near[i] 
        end # adjust orthogonal planes

        rec = Rectangle.new( near2, far2 )
        return rec
        
    end # of inset()
    
    def ins( v1, v2 )
        if v1 < v2
            return [ v1+@thickness, v2-@thickness ] 
        else
            return [ v2+@thickness, v1-@thickness ]
        end 
    end # ins()

end # of class Doughnut

class Rectangle
=begin
A Rectangle is created from a near corner and a far corner. 
The corners are [r,g,b] arrays. 

An orthogonal rectangle is created, if possible. 
The rectangle's @plane is one of 'rg', 'gb', 'br' or 'no' (Not Orthogonal). 

An orthogonal Rectangle's normal points toward the positive end of the
perpendicular axis. 

A non-orthogonal Rectangle's outside face will face the positive end 
of the blue axis, except that Rectangles parallel to the blue axis face the positive end of the green axis.
=end
    attr_reader :near, :far, :plane, :face, :group
    
    def initialize( near, far )
        
        r = 0; g = 1; b = 2;
        @near = near; @far = far
        
        if near[r] == far[r]
            @plane = 'gb'
        elsif near[g] == far[g]
            @plane = 'rb'
        elsif near[b] == far[b]
            @plane = 'rg'
        else
            @plane = 'no'
        end
        
    end # of initialize()

    def draw( *args ) # args == none or optional group
        corners = get_corners()
        if args.length == 1
            corners.insert( 0, args[0] )
        end

        @group = add_face_group( corners )
        @face = find_face( @group.entities )
        return @group
        
    end # of draw()
    
    def move( vec )
        @near[0] += vec[0]; @near[1] += vec[1]; @near[2] += vec[2]
        @far[0]  += vec[0]; @far[1]  += vec[1]; @far[2]  += vec[2]
        group.entities.erase_entities( group.entities.to_a() )
        draw()
    end # of move()
    
    def inspect()
        return 'Rectangle: @near=' + @near.inspect() + ', @far=' + @far.inspect() 
    end
    
    def get_corners()

        r = 0; g = 1; b = 2;
        ret = []
        ret[0] = @near
        ret[2] = @far
        if ( @plane == 'rg' ) || ( @plane == 'no' )
            ret[1] = [ @near[r], @far[g], @far[b] ]
            ret[3] = [ @far[r], @near[g], @near[b] ]
        elsif @plane == 'gb'
            ret[1] = [ @near[r], @near[g], @far[b] ]
            ret[3] = [ @near[r], @far[g], @near[b] ]
        else # @plane == 'rb'
            ret[1] = [ @near[r], @near[g] ,@far[b] ]
            ret[3] = [ @far[r], @near[g], @near[b] ]
        end

        return ret

    end # get_corners()
    
    def center()
        ret = [ 
            (@near[0] + @far[0]) / 2.0, 
            (@near[1] + @far[1]) / 2.0,
            (@near[2] + @far[2]) / 2.0 ]
        return ret
    end # of center()
    

end # of class Rectangle

class Timer
=begin
Written and placed in public domain by Martin Rinehart, January, 2010.

Thinking back to my Java days, I thought that UI.start/stop_timer() methods should really be replaced by a Timer class. You can do three things with a timer:
 
timer = Timer.new( pause_length, runnable )
timer.start()
timer.stop()

A Timer is initialized with a pause_length (seconds, including fractions of a second, often 1.0/frames_per_second) and a Runnable object. An object is Runnable if it has a run() method that returns a boolean (true==keep running, false==stop running). 

After the Timer's start() method is called, the Runnable's run() is called every pause_length seconds until the Timer's stop() method is called or the Runnable's run() method returns false.
=end

    # attr_reader :pause_length, :runnable, :running

    def initialize( pause_length, runnable )
        @pause_length = pause_length
        @runnable = runnable
        @running = false
    end

    def start()
        @id = UI.start_timer( @pause_length, true ) { 
            @running =  @runnable.run()
            stop() unless @running
        }
    end

    def stop()
        UI.stop_timer( @id )
        @running = false
    end

end # of class Timer

# test code

# class Mover

    # def initialize ( comp )
        # @comp = comp
    # end
    
    # def move( comp, r, g, b )
        # trans = comp.transformation().to_a
        # trans[12]+=r; trans[13]+=g; trans[14]+=b
        # comp.move!( Geom::Transformation.new(trans) )
        # Sketchup.active_model().active_view().invalidate()
    # end

# end # of class Mover

# class M1 < Mover
    # def run( frame )
        # move( @comp, -1, 1, 1 ) if frame < 84
        # move( @comp, 1, -1, -1 ) if frame >= 84 
    # end # of run()
    
# end # of class M1

# class M2 < Mover
    # def run( frame )
        # move( @comp, 0, 1, 1 ) if frame < 108
        # move( @comp, 0, -1, -1 ) if frame >= 108
    # end # of run()
    
# end # of class M1

# class M3 < Mover
    # def run( frame )
        # move( @comp, 1, 1, 1 ) if frame < 132
        # move( @comp,  -1, -1, -1 ) if frame >= 132
    # end # of run()
    
# end # of class M1

# model = Sketchup.active_model
# sels = model.selection
# c1 = sels[0]
# c2 = sels[1]
# c3 = sels[2]

# m1 = M1.new( c1 )
# m2 = M2.new( c2 )
# m3 = M3.new( c3 )

# Conductor.setup( 24, 8 )
# Conductor.register( m1, 1, 5 )
# Conductor.register( m2, 2, 6 )
# Conductor.register( m3, 3, 7 )
# begin
    # Conductor.start()
# rescue => e
    # puts e
# end

# end of sketchtalk_classes.rb