# sketch_talk_motion_classes.rb - classes for transforming geometry over time
# Copyright 2010, Martin Rinehart

class Action

    attr_reader :start, :stop, :values, :action_type

    # type enumeration
    Move = 1   unless defined?( Move )
    Rotate = 2 unless defined?( Rotate )
    Scale = 3  unless defined?( Scale )
    EyeTo = 4  unless defined?( EyeTo )
    EyeTw = 5  unless defined?( EyeTw )
    FovTo = 6  unless defined?( FovTo )
    FovTw = 7  unless defined?( FovTw )
    TarTo = 8  unless defined?( TarTo )
    TarTw = 9  unless defined?( TarTw )
    
    @@action_types = ['','Move','Rotate','Scale','EyeTo','EyeTw','FovTo','FovTw','TarTo','TarTw']
    
    def initialize( action_type, values, start, stop )
        @action_type = action_type
        @values = values
        @start = start
        @stop = stop
    end # of initialize()
    
    def action_name( action_number )
        return @@action_types[ action_number ]
    end # of action_name
    
    def camera_action?()
        return ( @action_type > 3 ) && ( @action_type < 10 )
    end
    
    def inspect()
        ret = '{Action: type=' + action_name( @action_type ) + 
            ', values=' + @values.inspect() + 
            ', seconds:' + @start.to_s() + ','+@stop.to_s() + '}'
        return ret
    end # of inspect()

end # of class Action

class Actor

    attr_reader :registrant
    
    def initialize( registrant )
        @registrant = registrant
        @comp_inst = @registrant.comp_inst()
        @action = @registrant.action()
        @action_type = @action.action_type()
        @values = @action.values()
        @transformation = nil
        
        frames = $sketch_talk_fps * ( @action.stop() - @action.start() )
        # $sketch_talk_fps is a float, so frames is a float
        
        if @action.camera_action?()
            cam = Sketchup.active_model().active_view().camera()
            
            case @action_type
                when Action::TarTo
                    dist = @values - cam.target().to_a()
                    @values = [ dist[0]/frames, dist[1]/frames, dist[2]/frames ]
                when Action::EyeTw
                    arr = @values.to_a()
                    @values = Geom::Vector3d.new( arr[0]/frames, arr[1]/frames, arr[2]/frames )
                when Action::FovTo
                    degrees = @values - cam.fov()
                    @values = degrees / frames
                when Action::TarTw
                    arr = @values.to_a()
                    @values = Geom::Vector3d.new( arr[0]/frames, arr[1]/frames, arr[2]/frames )
                when Action::EyeTo
                    dist = @values - cam.eye().to_a()
                    @values = [ dist[0]/frames, dist[1]/frames, dist[2]/frames ]
                when Action::FovTw
                    @values /= frames
            end # camera actions
        
        else # not camera_action?
        
            case @action_type
                when Action::Move
                    @values[0] /= frames # @values == [r, g, b]
                    @values[1] /= frames
                    @values[2] /= frames
                when Action::Rotate
                    # @values == [point, vector, degrees]
                    @transformation = Geom::Transformation.rotation(
                        @values[0], @values[1], (@values[ 2 ]*$radians_per_degree)/frames )
                when Action::Scale
                    # @values == [ point, rscale, gscale, bscale ]
                    pow = 1.0 / frames
                    rscale = @values[1]**pow
                    gscale = @values[2]**pow
                    bscale = @values[3]**pow
                    @transformation = 
                        Geom::Transformation.scaling( @values[0], rscale, gscale, bscale )
                else
                    raise 'Actor.new(): unknown action type: ' + @action_type.inspect()
            end
            
        end # camera_action?
    end
    
    def act( second, frame )
        if @action.camera_action?()

            cam = Sketchup.active_model().active_view().camera()

            case @action_type
                when Action::TarTo
                    cam.set( cam.eye, cam.target + @values, [0,0,1] )
                when Action::TarTw
                    cam.set( cam.eye, cam.target + @values, [0,0,1] )
                when Action::EyeTo
                    cam.set( cam.eye + @values, cam.target, [0,0,1] )
                when Action::EyeTw
                    cam.set( cam.eye + @values, cam.target, [0,0,1] )
                when Action::FovTo
                    cam.fov += @values
                when Action::FovTw
                    cam.fov += @values
            end
            
        else # not camera action
        
            trans_was = @comp_inst.transformation()
            case @action_type
                when Action::Move then
                    arr = trans_was.to_a()
                    for i in 12..14 do arr[i] += @values[i-12] end
                    trans_is = Geom::Transformation.new( arr )
                when Action::Rotate then
                    trans_is = trans_was * @transformation
                when Action::Scale then
                    trans_is = trans_was * @transformation
                else
                    raise 'Actor.act(): unknown action type: ' + @action_type.inspect()
            end
            @registrant.comp_inst().move!( trans_is ) 
            
        end # camera action?
    end
    
    def inspect()
        ret = '{Actor: ' + @registrant.inspect()
        ret += ', ' + @comp_inst.transformation().to_a().inspect() unless 
            @comp_inst.is_a?( String )
        ret += '}'
        return ret
    end # of inspect()
    
end # of class Actor

class Call
    attr_reader :actors
    attr_accessor :stop_time

    def initialize( stop_time, actors )
        @stop_time = stop_time
        @actors = actors
    end
    
    def call( second, frame )

        for actor in @actors 
            actor.act( second, frame )
        end
        
    end # of call()
    
    def inspect() 
        ret = '{Call: stop=' + @stop_time.to_s() +
            ' Actors: '
        for actor in @actors do ret += actor.registrant.name + ',' end
        ret = ret.chop + '}'
        return ret
    end # of inspect()
    
end # of class Calls

class Conductor
=begin
The Conductor becomes the view's animation. His first job is to add registrant_list that wish to move (translate, rotate, scale) to his list of registrant_list. The "go" command signals that registration is over and it's time to get organized and start moving things. 

The conductors first task after "go" is to get organized. From the list of registrant_list he creates "Actors" (registrant plus an "act() method). He creates lists of Actors to call for each time period. Once organized, the conductor passes himself to the active_view's animation.

The view then begins calling the Conductor for each new frame. During each frame, every Actor that wanted a call is called. What the Actors act() methods do is unknown to the Conductor. Typically they will move, rotate and scale.

Organization is non-trivial. The Conductor organizes the Actors into an array of Call objects. The Call object holds a stop time and a list of registrant_list to call. It also has a call() method the conductor will call until its stop time is reached. The call() method calls the individual Actors. On reaching the Call's stop time, the Conductor moves on to the next Call object.

The goal here is to do the hard work before beginning the animation. Then call the appropriate registrant_list with minimal overhead. There are 40+ millis per frame. We want to leave as many millis as possible for the registrants to do their thing(s).
=end

    def initialize()
        @@registrant_list = [] # Array of Registrant objects.
        @@actor_list = [] # Array of Actor objects.
        @@call_list = [] # Array of Call objects.

        @@num_reg = 0
        @@frame_num = -1 
        @@delay = 1.0 / ( $sketch_talk_fps + 6 ) # Guesstimated frame delay.
        @@second = -1
        @@frames_per_sec = $sketch_talk_fps.to_i()
        @@call_ptr = 0
        @@start_time = Time.now()
        @@paused = false
        @@paused_time = nil
        
        @@player_ready = false
        @@ready_to_go = false
        
        @@wd = UI::WebDialog.new( 'Video Player', true, 'video_player', 
            180, 70, 100, 100, true )
        @@wd.set_file( Sketchup.find_support_file('video_player.html',
            'Plugins/ruby_console_pro/') )

        @@wd.add_action_callback( 'play' ) do |dlg, msg|
            case msg
                when 'hello'
                    @@player_ready = true
                    if @@ready_to_go
                        @@wd.execute_script( 'go_black()' )
                        @@start_time = Time.now() # RIGHT HERE!
                        Sketchup.active_model().active_view().animation = self
                    end
                when 'pause' then pause()
                when 'resume' then resume()
            end
        end # of play()
        
    end # of initialize()
    
    def pause()
    
        @@paused = true
        @@paused_time = Time.now()
    
    end # of pause()
    
    def resume()

        # Adjust start time because conductor uses it to adjust speed.
        pause_length = Time.now() - @@paused_time
        @@start_time += pause_length
        @@paused = false
        
        Sketchup.active_model().active_view().animation = self
    
    end # of resume()
    
# -------------------- The Animation Interface: -------------------- 

    # Called by the active_view, frequently.
    def nextFrame( view )
        
        @@frame_num += 1
        if ( @@frame_num % @@frames_per_sec ) == 0
            @@second += 1
            adjust_speed()
            if @@call_list[ @@call_ptr ].stop_time == @@second
# for call in @@call_list do outs( call.inspect() ) end
                @@call_ptr += 1
            end
        end
        
        if @@call_ptr == @@call_list.length()
            outs Time.now() - @@start_time
            @@wd.close()
            $sketch_talk_conductor = nil
            return false 
        end
        
        @@call_list[@@call_ptr].call( @@second, @@frame_num )
        
        view.show_frame( @@delay )
        
        return !@@paused # returned false (above) if @@call_list was exhausted
        
    end # of nextFrame()
    
    # def stop()
        # any cleanup? doubt it, but just in case
    # end # of stop()
    
# -------------------- Other Public Methods: --------------------

    # Register a ComponentInstance with move vector and move times.
    def register( comp_inst, action )
    
        name = 'R' + @@num_reg.to_s()
        reg = Registrant.new( comp_inst, name, action )
        @@registrant_list.push( reg )
        @@num_reg += 1

        return true
        
    end # of register()
    
    # Called by the "go" command function.
    def start()
# outs 'starting'
# for reg in @@registrant_list do outs( reg.inspect() ) end
# outs
        for reg in @@registrant_list do @@actor_list.push( Actor.new(reg) ) end
# for actor in @@actor_list do outs( actor.inspect() ) end
# outs
        organize()
# for call in @@call_list do outs( call.inspect() ) end
# outs
        @@ready_to_go = true
        @@wd.show()
        @@wd.execute_script( 'go_black()' )
        
    end # of start()
    
# -------------------- Private Methods --------------------
    
    private
    
    # Attempt to go at correct fps, regardless of machine speed.
    def adjust_speed()
        return if @@second == 0
        time_elapsed = Time.now() - @@start_time
        time_should_be = @@second

        @@delay = ( @@delay * time_should_be ) / time_elapsed
# outs 'elapsed: ' + time_elapsed.to_s + ', should_be: ' + time_should_be.to_s + ', delay:' + @@delay.to_s()
    end # of adjust_speed()
    
    # Converts array of Actor objects to Array of Call objects.
    def organize() 
        call_ptr = 0
        
        # There is no guarantee that any Registrant starts at 0.
        @@call_list.push( Call.new(0, []) )
        
        # Registrant stop values are the second at which motion stops.
        # If one stops at 10, it will not be in the call list for second 10.
        last_sec = find_last() - 1
        old_list = []
        
        for sec in 0..last_sec
        
            # Build list of Actor objects to call.
            list = []
            for actor in @@actor_list
                list.push( actor ) if actor.registrant.wants_call?( sec ) 
            end
            
            next_sec = sec+1
            if list == old_list
                # If list is unchanged, increment its stop time.
                @@call_list[ call_ptr ].stop_time = next_sec
            else
                # Otherwise, push a new Call object onto array.
                @@call_list.push( Call.new(next_sec, list) )
                call_ptr += 1
                old_list = list
            end
            
        end # seconds loop
        
    end # of organize()

# -------------------------- support functions ------------------------

    # Returns the last "stop" value in the array of Registrants.
    def find_last()
        last = 0
        for reg in @@registrant_list do last = [ last, reg.action.stop ].max end
        return last
    end
    
end # of class Conductor

class Registrant

    attr_reader :comp_inst, :name, :action
    attr_accessor :comp_inst

    def initialize( comp_inst, name, action )
        @comp_inst = comp_inst
        @name = name
        @action = action
    end # of initialize()
    
    def wants_call?( sec )
        return ( sec >= @action.start ) && ( sec < @action.stop )
    end
    
    def inspect()
        return '{Registrant: name=' + @name + 
            ', action=' + @action.inspect() + '}'
    end # of inspect()
    
end # of class Registrant

# end of sketch_talk_motion_classes.rb